// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

#pragma once

#include <memory>
#include <string>
#include <vector>

#include "arrow/status.h"
#include "arrow/testing/visibility.h"

namespace arrow::util {

class ARROW_TESTING_EXPORT Process {
 public:
  Process();
  ~Process();

  Status SetExecutable(const std::string& path);
  void SetArgs(const std::vector<std::string>& args);
  void SetEnv(const std::string& name, const std::string& value);
  void IgnoreStderr();
  Status Execute();
  bool IsRunning();
  uint64_t pid();

 private:
  class Impl;
  std::unique_ptr<Impl> impl_;
};
}  // namespace arrow::util
